/*
 *	eZ80 Processor Module
 *	Brandon Wilson
 *	2015
 */

#include "ez80.hpp"

static const char *const condNames[] =
{
  "nz",
  "z",
  "nc",
  "c",
  "po",
  "pe",
  "p",
  "m"
};

static void OutReg(int rgnum)
{
  out_register(ph.regNames[rgnum]);
}

bool idaapi outop(op_t &x)
{
  if ( !x.showed() ) return 0;

  switch ( x.type )
  {
  case o_cond:

        if ( x.Cond == oc_not ) return 0;
        {
          char buf[3];
          qstrncpy(buf, condNames[ x.Cond ], sizeof(buf));
          out_keyword(buf);
        }
        break;

  case o_reg:

        OutReg(x.reg);
        break;

  case o_displ:
	  {
        out_symbol('(');
        OutReg(x.phrase);
        char buf[MAXSTR];
        if ( isOff(uFlag,x.n)
          && get_offset_expression(cmd.ea,x.n,cmd.ea+x.offb,x.addr,buf,sizeof(buf)) )
        {
          out_symbol('+');
          OutLine(buf);
        }
        else
        {
          int offbit = OOFW_8;
          OutValue(x,OOF_ADDR|offbit|OOFS_NEEDSIGN|OOF_SIGNED);
        }
        out_symbol(')');
        break;
	  }
  case o_phrase:
	  {
          out_symbol('(');
          OutReg(x.phrase);
          out_symbol(')');
        break;
	  }
  case o_void:
        return 0;
  case o_imm:
        {
          const char *name = NULL;
          bool needbrace = false;
            switch ( cmd.itype )
            {
              case Z80_im:
              case Z80_bit:
              case Z80_res:
              case Z80_set:
//                name = z80_find_ioport_bit(x.value);
                break;
              case Z80_in0:
              case Z80_out0:
                name = z80_find_ioport(x.value);
              case Z80_in:
              case Z80_out:
              case Z80_outaw:
              case Z80_inaw:
                out_symbol('(');
                needbrace = true;
                break;
              default:
                break;
            }
          if ( name != NULL )
            out_line(name, COLOR_IMPNAME);
          else
		  {
            OutValue(x, 0);
		  }
          if ( needbrace ) out_symbol(')');
        }
        break;

  case o_mem:
        out_symbol('(');
  case o_far:
  case o_near:
        {
          ea_t base;
          if ( x.type == o_mem )
            base = dataSeg_op(x.n);
          else
            base = codeSeg(x.addr,x.n);
          ea_t v = toEA(base,x.addr);
//          const char *ptr;
          if ( v == cmd.ea && ash.a_curip != NULL )
          {
            OutLine(ash.a_curip);
          }
          else if ( !out_name_expr(x, v, x.addr) )
          {
            out_tagon(COLOR_ERROR);
            OutLong(x.addr,16);
            out_tagoff(COLOR_ERROR);
            QueueMark(Q_noName,cmd.ea);
          }
          if ( x.type == o_mem )
            out_symbol(')');
        }
        break;

  default:
        warning("bad optype %x", x.type);
        break;
  }
  return 1;
}

static bool isIxyByte(op_t &x)
{
  return x.type == o_reg &&
        (x.reg == R_xl ||
         x.reg == R_xh ||
         x.reg == R_yl ||
         x.reg == R_yh);
}

//----------------------------------------------------------------------
inline bool isIxyOperand(op_t &x)
{
  return isIxyByte(x) || x.type == o_displ;
}

void idaapi out(void)
{
  char buf[MAXSTR];
  char* postfix = NULL;

  init_output_buffer(buf, sizeof(buf));

  //I'm not so sure about this...
  if ((isIxyByte(cmd.Op1)  || isIxyByte(cmd.Op2)  ||
                 (cmd.itype == Z80_adc || cmd.itype == Z80_sbc) &&
                 (cmd.Op1.reg == R_ix || cmd.Op1.reg == R_iy)))
  {
    OutBadInstruction();
  }

  switch (cmd.auxpref)
  {
  case 0x40:
	  postfix = ".sis";
	  break;
  case 0x49:
	  postfix = ".lis";
	  break;
  case 0x52:
	  postfix = ".sil";
	  break;
  case 0x5B:
	  postfix = ".lil";
	  break;
  default:
	  break;
  }
  OutMnem(8, postfix);

  char comma = (char)out_one_operand(0);

  if ( comma && cmd.Op2.showed() && cmd.Op2.type != o_void ) {
    out_symbol(',');
    out_symbol(' ');
  }

  out_one_operand(1);

  if ( isVoid(cmd.ea,uFlag,0) ) OutImmChar(cmd.Op1);
  if ( isVoid(cmd.ea,uFlag,1) ) OutImmChar(cmd.Op2);

  term_output_buffer();
  gl_comm = 1;
  MakeLine(buf);
}

void idaapi ez80_segstart(ea_t ea)
{
  segment_t *segm = getseg(ea);

  char sname[MAXNAMELEN];
  get_true_segm_name(segm, sname, sizeof(sname));

  if ( inf.s_org )
  {
    ea_t org = ea - get_segm_base(segm);
    if ( org != 0 )
    {
      char buf[MAX_NUMBUF];
      btoa(buf, sizeof(buf), org);
      printf_line(inf.indent, COLSTR("%s %s", SCOLOR_ASMDIR), ash.origin, buf);
    }
  }
}

void idaapi ez80_header(void)
{
  char buf[MAXSTR];
  gen_cmt_line("Processor       : %s [%s]", device[0] ? device : inf.get_proc_name(buf), deviceparams);
  //gen_cmt_line("Target assembler: %s",ash.name);
  //if ( ash.header != NULL )
  //  for ( const char **ptr=ash.header; *ptr != NULL; ptr++ ) MakeLine(*ptr,0);
}

void idaapi ez80_footer(void)
{
  char buf[MAXSTR];
  char *const end = buf + sizeof(buf);
  if ( ash.end != NULL )
  {
    MakeNull();
    char *ptr = tag_addstr(buf, end, COLOR_ASMDIR, ash.end);

    char name[MAXSTR];
    if ( get_colored_name(BADADDR, inf.beginEA, name, sizeof(name)) != NULL )
    {
      MakeLine(buf, inf.indent);
      ptr = buf;
      APPEND(ptr, end, ash.cmnt);
      APPCHAR(ptr, end, ' ');
      APPEND(ptr, end, name);
    }
    MakeLine(buf, inf.indent);
  }
  else
  {
    gen_cmt_line("end of file");
  }
}
